/*
	"pad.c" - Sample Code

	Copyright (C) 1998  Sony Computer Entertainment Europe
	All rights Reserved
*/

	// This file contains all source concerning the controller pad.

/* Basically, the constraints on this demo meant that the Dual Shock
	was meant to be locked in Analog Mode. It is very helpful for many
	functions to have a simple flag to check to let them know if:
		A) There is a valid controller plugged in.
		B) If that controller is in analog mode.
	Two such variables exist - validControllerFlag and analogControllerFlag.
*/

#define _PAD_
#include <stdio.h>
#include <sys/types.h>
#include <libetc.h>
#include <libpad.h>
#include "pad.h"

//#define DEBUG

	// When developing, it is very useful to have the ability to turn
	// vibration off, which prevents your Dual Shock having a fit and falling
	// off the desk!
#define VIBRATION

	// This variable contains the intensity of Motor1. It is an integer instead
	// of a u_char because it will temporarily contain the sum of two different
	// factors (the vibration caused by the stomp and the vibration caused by
	// the volcano), and this will have to be clamped at 255 (maximum intensity)
	// before we send it to the controller.
	// The vibration of the controller due to the 'stomp' is stored in this
	// variable. The vibration due to the volcano is stored in another variable,
	// and the two are combined in the function processController.

static int StompDualShockIntensity = 0;

	// Controller Information buffer - the libraries fill this with information
	// from the controller like the buttons status and the position of the
	// analog controllers.
static unsigned char padbuff[2][34];

	// This alignment buffer is a constant array that we use to align data
	// sent to the Dual Shock. This array won't change in the near future.
static unsigned char align[6]={0,1,0xFF,0xFF,0xFF,0xFF};

	// This array holds the two values which specify the motor vibration
	// intensity.
unsigned char MotorStatus[2];

	// The following strings and lines were used to display the state of
	// the controller on screen. A ring buffer is used to store changes
	// in the state. This is useful for debugging. This is not included in
	// the final release.

#ifdef DEVKIT
static char* messages[] = {
	"NO CONTROLLER",
	"FINDING PAD",
	"CHECKING FOR NON DSCONTROLLERS",
	"GETTING ACTUATOR INFORMATION",
	"COMMUNICATING WITH CONTROLLER",
	"COMMUNCATION COMPLETE - STABLE",
	"UNKNOWN STATE!!!!!!!!" };

#define MESSAGE_BUFF_SIZE 10
signed char messring[MESSAGE_BUFF_SIZE] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1 };

#endif

// ========================================================================
		// Init controller initialises the pad buffers and attempts to set
		// up the alignment.

void initController(void) {
	PADvalidState_flag = FALSE;
	PADanalogController_flag = FALSE;

	MotorStatus[0] = MotorStatus[1] = 0;

	PadInitDirect(&padbuff[0][0],&padbuff[1][0]);
	PadStartCom();
	VSync(20);

		// This says that the vibration info for controller 0 is to be found
		// at the address of MotorStatus[0], and is 2 bytes long (1 for Motor0,
		// 1 for Motor1);
	PadSetAct(0, &MotorStatus[0], 2);
}

// ========================================================================

		// This will stop vibration by setting the motor values to 0 and
		// waiting for 2 VSyncs before shutting down the controller. We
		// wait 2 VSyncs to make sure the data is transmitted to the controller.

void stopController(void) {
	MotorStatus[0] = MotorStatus[1] = 0;
	VSync(2);
	PadStopCom();
}

// ========================================================================

		// This function returns the state of the buttons for Pad 0.
u_short getPadButtons() {
	if(PADvalidState_flag) {
		return ~((padbuff[0][2] << 8) | padbuff[0][3]);
	}
	return 0;	// No buttons pressed.
}

// ========================================================================

	// This function returns 4 u_char values which indicate the state of
	// the analog sticks. It 'fakes' values for non-analog controllers.
void getAnalogControllerValues(u_char *vals) {
	u_long pad;
	int i;
		// 'normalPadDelta' is the "fake" analog value we use where there is no
		// analog pad, and 'threshhold' is the difference from "Stick rest" we
		// require in an analog stick before it actually takes effect.

		// A threshold is required because the analog controller can oscillate between
		// two appropriate values.

		// For this demo, we don't want to center the right analog controller,
		// which will control the orientation of the head, so we don't do
		// any thresholding.
	const int normalPadDelta = 0x10, threshhold = 0x20;

	if(PADvalidState_flag) {
		if(PADanalogController_flag) {
			for(i = 0; i < 2; i++) { // Do right hand controller
				vals[i] = ((int)padbuff[0][i+4] - 0x80) / 5 + 0x80;
			}
			for(i = 2; i < 4; i++) {	// Do left hand ones
				if(padbuff[0][i+4] > (0x80 + threshhold) || padbuff[0][i+4] < (0x80 - threshhold)) {
					vals[i] = ((int)padbuff[0][i+4] - 0x80) / 5 + 0x80;
				} else {
					vals[i] = 0x80;
				}
			}
		} else {
				// If there's no analog, we fake the values using the normal
				// controller pad. Obviously for the right stick we cannot
				// do this, so digital controllers cannot control the dino head.

			pad = ~((padbuff[0][2] << 8) | padbuff[0][3]);
			if(pad & PADLleft ) {
				vals[2] = 0x80 - normalPadDelta;
			} else if(pad & PADLright) {
				vals[2] = 0x80 + normalPadDelta;
			} else {
				vals[2] = 0x80;
			}
			if(pad & PADLdown ) {
				vals[3] = 0x80 + normalPadDelta;
			} else if(pad & PADLup   ) {
				vals[3] = 0x80 - normalPadDelta;
			} else {
				vals[3] = 0x80;
			}
			vals[0] = 0x80;
			vals[1] = 0x80;
		}
	} else {
		vals[0] = vals[1] = vals[2] = vals[3] = 0x80;	// No controller = no movement
	}
#ifdef DEBUG
	FntPrint("%02x %02x %02x %02x\n",vals[0],vals[1],vals[2],vals[3]);
#endif
}

// ========================================================================

	// Initiates a shudder in the controller.
void doShock(void) {
	StompDualShockIntensity = 255;
}

// ========================================================================

	// processController is the function called every frame that performs
	// a number of 'administrative' functions. It checks to see that
	// controllers are connected, and updates the flags. If we are debugging,
	// it updates the ring buffer which hold the last 10 states.
	// If a Dual Shock controller is inserted, it is locked into analog mode.

void processController(void) {

	int state,i,numIDs,tabID;

/* Our Dual Shock needs two functions performed before it is ready
	for use. Firstly, the PadSetActAlign function must be called (which
	sets up the Dual Shock to receive data from the PlayStation), and
	then it is locked into analog mode (this is optional). However,
	calling PadSetActAlign causes the controller to change to a state
	where you cannot lock the controller for a VSync! So we have to
	perform the lock on the next time processController is called. These
	flags keep track of what has been done already (PadSetActAlign/Lock)
	so we know what to attempt suring the current processController call.
*/
	static bool setActAlign_flag = FALSE,lockControllerCheck_flag = FALSE;

		// This variable contains the desired intensity of the motor due
		// to the volcano.
	extern int VolcanoDualShockIntensity;


#ifdef DEVKIT
		// Ring buffer stuff (see above)
	state = PadGetState(0);
	switch(state) {
		case PadStateDiscon:
			k = 0;
			break;
		case PadStateFindPad:
			k = 1;
			break;
		case PadStateFindCTP1:
			k = 2;
			break;
		case PadStateReqInfo:
			k = 3;
			break;
		case PadStateExecCmd:
			k = 4;
			break;
		case PadStateStable:
			k = 5;
			break;
		default:
			k = 6;
			break;
	}
	if(messring[m] != k) {
		m++;
		m %= MESSAGE_BUFF_SIZE;
		messring[m] = k;
	}

		// Set this to '#ifdef 1' to see the last 10 controller states.
#if 0
	j = (m + 1) % MESSAGE_BUFF_SIZE;
	for(i = 0; i < MESSAGE_BUFF_SIZE; i++) {
		FntPrint("%d - %s\n",i,(messring[j] == -1)?"":messages[messring[j]]);
		j++;
		j %= MESSAGE_BUFF_SIZE;
	}

	FntPrint("\n\n\n%d\n",PadInfoMode(0,InfoModeCurID,0));
#endif

#endif

		// Process controller 0 only
	state = PadGetState(0);

		// If the state is PadStateStable (i.e. Dual Shock), or PadStateFindCTP1,
		// then we have a valid controller and it's button/analog values will be valid.
	if((state==PadStateStable) || (state==PadStateFindCTP1)) {

			// This flag will be false when the state was not previously stable.
		if(PADvalidState_flag == FALSE) {

				// If we haven't set the alignment yet...
			if(setActAlign_flag == FALSE) {
				if(state==PadStateStable) {	// Dual shock
					if(!PadSetActAlign(0, align)) {
						printf("PadSetActAlign failed in processController().\n");
					} else {
						setActAlign_flag = TRUE;
					}
				}
			} else if(lockControllerCheck_flag == FALSE) {
					// Search the table of ID's for an analog controller
				numIDs = PadInfoMode(0,InfoModeIdTable,-1);
				while(--numIDs >= 0) {
					tabID = PadInfoMode(0,InfoModeIdTable,numIDs);
					if(tabID == 5  ||  tabID == 7)
						break;
				}

				if(numIDs < 0) {	// Analog not found
					lockControllerCheck_flag = TRUE;
				} else {
						// Analog found, so lock it as such.
					PADanalogController_flag = TRUE;
						// If it's not already in analog mode...
					if(PadInfoMode(0,InfoModeCurExOffs,0) != numIDs) {
						if(PadGetState(0) == PadStateStable) {
							if(!PadSetMainMode(0,numIDs,3)) {	// Change & lock it
							// Lock failed!
								printf("Lock controller state failed!\n");
								PADvalidState_flag = FALSE;
								PADanalogController_flag = FALSE;
							} else {
								printf("Lock controller state succeeded.\n");
								lockControllerCheck_flag = TRUE;
							}
						} else {
							PADanalogController_flag = FALSE;
							lockControllerCheck_flag = TRUE;
						}
					} else {
						PADanalogController_flag = TRUE;
						lockControllerCheck_flag = TRUE;
					}
				}
			}

			PADvalidState_flag = lockControllerCheck_flag && setActAlign_flag;
		}

			// We have to handle the old style analog controller here,
			// because we can't select/lock the mode on it.

				// If state is stable and not a Dual Shock...
		if(state==PadStateFindCTP1) {
			PADvalidState_flag = TRUE;
			i = PadInfoMode(0,InfoModeCurID,0);
			if(i == 5  ||  i == 7) {
				PADanalogController_flag = TRUE;
			} else {
				PADanalogController_flag = FALSE;
			}
		}

	} else {
		PADvalidState_flag = FALSE;
		PADanalogController_flag = FALSE;
		if(state != PadStateReqInfo) {
			setActAlign_flag = FALSE;
			lockControllerCheck_flag = FALSE;
		}
	}

#ifdef DEBUG
	if(PADvalidState_flag) FntPrint("Valid state.\n");
	if(PADanalogController_flag) FntPrint("Analog Controller.\n");
#endif

		// Different variances in motor intensity over time can produce
		// different physical effects. This piece of code reduces the
		// intensity with respect to a very simple piece-wise function.
		// This was achieved with trial and error. It was designed to simulate
		// a realistic 'stomp' rumble.

	if(StompDualShockIntensity >= 200) {
		StompDualShockIntensity -= 10;
	} else if(StompDualShockIntensity >= 100) {
		StompDualShockIntensity -= 15;
	} else if(StompDualShockIntensity >= 20) {
		StompDualShockIntensity -= 20;
	} else {
		StompDualShockIntensity = 0;
	}

		// This puts the vibration into the area which is transmitted
		// to the controller. Because 2 values are combined we have
		// to clamp them at 255.
#ifdef VIBRATION
	if(StompDualShockIntensity + VolcanoDualShockIntensity > 255) {
		MotorStatus[1] = 255;
	} else {
		MotorStatus[1] = StompDualShockIntensity + VolcanoDualShockIntensity;
	}
#endif
}

